import torch
import torch.nn as nn
import numpy as np
import os

class ModularBaseCNN(nn.Module):
    device=torch.device("cuda"if torch.cuda.is_available() else "cpu")

    def __init__(self,input_shape,num_target_classes, num_experts):
        super(ModularBaseCNN,self).__init__()
        self.depth=64
        self.dense_depth=4096
        self.validation=[]
        self.max_epoch=5
        self.input_shape=input_shape
        self.num_target_classes=num_target_classes
        self.CEloss=nn.CrossEntropyLoss()
        self.num_experts=num_experts
        self.expertBase=nn.ModuleList([self.getExpertBase() for i in range(self.num_experts)])
        self.expertDense=nn.ModuleList([self.getExpertDense() for i in range(self.num_experts)])
        self.gatingBase=self.getGatingBase()
        self.gatingDense=self.getGatingDense()
        self.optimizer=torch.optim.Adam(self.parameters(),float('1.0e-5'))
 
    def getGatingBase(self):
        return nn.Sequential(
            nn.Conv2d(1,self.depth,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth),
            nn.PReLU(),
            nn.Conv2d(self.depth,self.depth,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth),
            nn.PReLU(),
            nn.MaxPool2d(2),

            nn.Conv2d(self.depth, self.depth*2, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*2),
            nn.PReLU(),
            nn.Conv2d(self.depth*2, self.depth*2, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*2),
            nn.PReLU(),
            nn.MaxPool2d(2),

            nn.Conv2d(self.depth*2, self.depth*4, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*4),
            nn.PReLU(),
            nn.Conv2d(self.depth*4, self.depth * 4, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*4),
            nn.PReLU(),
            nn.Conv2d(self.depth*4, self.depth * 4, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*4),
            nn.PReLU(),
            nn.MaxPool2d(2))

    def getGatingDense(self):
         return nn.Sequential(
            nn.Linear(self.depth*8*8*4,self.dense_depth),
            nn.PReLU(),
            nn.Linear(self.dense_depth,self.dense_depth),
            nn.PReLU(),
            nn.Linear(self.dense_depth,self.num_experts)
        )
    def getExpertBase(self):
        return nn.Sequential(
            nn.Conv2d(self.depth*4,self.depth*8,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.Conv2d(self.depth*8,self.depth*8,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.Conv2d(self.depth*8, self.depth*8, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.MaxPool2d(2),
            
            nn.Conv2d(self.depth*8,self.depth*8,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.Conv2d(self.depth*8,self.depth*8,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.Conv2d(self.depth*8,self.depth*8,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.MaxPool2d(2))
    
    def getExpertDense(self):
        return nn.Sequential(
            nn.Linear(self.depth*2*2*8,self.dense_depth),
            nn.PReLU(),
            nn.Linear(self.dense_depth,self.dense_depth),
            nn.PReLU(),
            nn.Linear(self.dense_depth,self.num_target_classes)
        ) 


    def f(self, x):
        x = x.cuda(self.device)
       # x = self.Expert_model(x.reshape((-1, 1, self.input_shape[0], self.input_shape[1])))
        
        #x = self.dense_model(x.view(x.size(0), -1))
        x=self.gatingBase(x.reshape((-1,1,self.input_shape[0],self.input_shape[1])))
        gates=self.gatingDense(x.view(x.size(0),-1))
        #print(gates.size())
        pred=torch.zeros((x.size(0),self.num_target_classes),dtype=torch.float).cuda(self.device)   
        for i in range(self.num_experts):
            out = self.expertBase[i](x)
            ##print(x.size())
            out = self.expertDense[i](out.view(out.size(0), -1))
            #print(out.size())
            gate_norm=(torch.exp(gates[:,i])/torch.sum(torch.exp(gates),dim=1))
            #print(gate_norm.size())
            pred+=out*gate_norm[:,None]        

        return torch.softmax(pred, dim=1)
    # 1. Fix y so that we get a vector output... on everything, then redo all previous tests
    # 2. If 1. didn't work, use already implemented logsumexp
    # 3. If 2. didn't work, try using a squashing function on out to prevent y-out from becoming too large

    def loss(self,x,y):
        x=torch.tensor(x,dtype=torch.float).cuda(self.device)
        x=self.gatingBase(x.reshape((-1,1,self.input_shape[0],self.input_shape[1])))
        true=torch.zeros((x.size()[0],self.num_target_classes),device=self.device)
        for i, target in enumerate(y):
            true[i,int(target)]=1.
        gates=self.gatingDense(x.view(x.size(0),-1))
        res=torch.zeros(x.size()[0],dtype=torch.float,device=self.device)
        shift=0
        for i in range(self.num_experts):
            out = self.expertBase[i](x)
            out = self.expertDense[i](out.reshape(out.size(0), -1))
            gate_norm=torch.exp(gates[:,i])/torch.sum(torch.exp(gates))
            if shift==0:
                shift=torch.max(0.5*(true-out)*(true-out))
            distance=(-0.5*torch.norm((true-out)*(true-out),dim=1))-shift
            tmp=gate_norm*torch.exp(distance)
            res=res+tmp
        loss=torch.sum(shift-torch.log(res))
        return loss 

    def accuracy(self,gen_test_batch,forgiveness=1):
        torch.cuda.empty_cache()
        accuracy=[]
        print('Calculating accuracy')
        for data,target in gen_test_batch:
            data=torch.tensor(data,dtype=torch.float).cuda(self.device)
            target = torch.tensor(target, dtype=torch.float).cuda(self.device)
            accuracy.append(torch.mean(torch.tensor(
                [target[i] in batch_item for i,batch_item in enumerate(self.f(data).topk(k=forgiveness,dim=1).indices)],
                dtype=torch.float
            )))
        out = np.average(accuracy)
        print('Current accuracy %f'%out)
        return out

    def load(self):
        if os.path.isfile('Models/MixtureOfExperts_M1_state.pt'):
            state = torch.load('Models/MixtureOfExperts_M1_state.pt')
            self.load_state_dict(state['model_state'])
            self.optimizer.load_state_dict(state['optimizer_state'])

    def save(self):
        print("Saving...")
        self.to(torch.device('cpu'))
        self.optimizer.zero_grad()
        torch.save({
            'model_state': self.state_dict(),
            'optimizer_state': self.optimizer.state_dict()
        }, 'Models/MixtureOfExperts_M1_state.pt')
        self.to(self.device)

    @staticmethod
    def clear():
        print("Clearing unused variables")
        torch.cuda.empty_cache()


    def trainer(self, gen_training_batch,gen_test_batch,batch_size):
        print("Training on %s"%self.device) 
        loss=np.zeros(self.max_epoch)
        accuracy=np.zeros(self.max_epoch)
        for epoch in range(self.max_epoch):
            # Loss
            loss_epoch=[]
            print('Starting training epoch %d'%epoch)
            for data, target in gen_training_batch[epoch]:
                self.optimizer.zero_grad()
                current_loss =self.loss(data.reshape((batch_size,1,self.input_shape[0],self.input_shape[1])),target)
                current_loss.backward()
                self.optimizer.step()
                loss_epoch.append(current_loss.item())
            
            accuracy[epoch]=self.accuracy(gen_test_batch[epoch])
            loss[epoch]=np.average(loss_epoch)
            print('Finishing trainng: Average loss: %f\n'%loss[epoch])
        return loss,accuracy


