import torch
import torch.nn as nn
import numpy as np
import os

class ModularNCLCNN(nn.Module):
    device=torch.device("cuda"if torch.cuda.is_available() else "cpu")

    def __init__(self,input_shape,num_target_classes, num_experts):
        super(ModularNCLCNN, self).__init__()
        self.depth=64
        self.dense_depth=4096
        self.penalty_strength=1
        self.validation=[]
        self.max_epoch=5
        self.input_shape=input_shape
        self.num_target_classes=num_target_classes
        self.CEloss=nn.CrossEntropyLoss()
        self.num_experts=num_experts
        self.expertBase=nn.ModuleList([self.get_expert_base() for i in range(self.num_experts)])
        self.expertDense=nn.ModuleList([self.get_expert_dense() for i in range(self.num_experts)])
        self.optimizer=torch.optim.Adam(self.parameters(),float('1.0e-5'))
 
    def get_expert_base(self):
        return nn.Sequential(
            nn.Conv2d(1,self.depth,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth),
            nn.PReLU(),
            nn.Conv2d(self.depth,self.depth,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth),
            nn.PReLU(),
            nn.MaxPool2d(2),

            nn.Conv2d(self.depth, self.depth*2, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*2),
            nn.PReLU(),
            nn.Conv2d(self.depth*2, self.depth*2, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*2),
            nn.PReLU(),
            nn.MaxPool2d(2),

            nn.Conv2d(self.depth*2, self.depth*4, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*4),
            nn.PReLU(),
            nn.Conv2d(self.depth*4, self.depth * 4, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*4),
            nn.PReLU(),
            nn.Conv2d(self.depth*4, self.depth * 4, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*4),
            nn.PReLU(),
            nn.MaxPool2d(2),

            nn.Conv2d(self.depth*4,self.depth*8,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.Conv2d(self.depth*8,self.depth*8,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.Conv2d(self.depth*8, self.depth*8, kernel_size=3, padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.MaxPool2d(2),

            nn.Conv2d(self.depth*8,self.depth*8,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.Conv2d(self.depth*8,self.depth*8,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.Conv2d(self.depth*8,self.depth*8,kernel_size=3,padding=1),
            nn.BatchNorm2d(self.depth*8),
            nn.PReLU(),
            nn.MaxPool2d(2))
    
    def get_expert_dense(self):
        return nn.Sequential(
            nn.Linear(self.depth*2*2*8,self.dense_depth),
            nn.PReLU(),
            nn.Linear(self.dense_depth,self.dense_depth),
            nn.PReLU(),
            nn.Linear(self.dense_depth,self.num_target_classes)
        ) 


    def f(self, x):
        x = x.cuda(self.device)
       # x = self.Expert_model(x.reshape((-1, 1, self.input_shape[0], self.input_shape[1])))
        
        #x = self.dense_model(x.view(x.size(0), -1))
        pred=torch.zeros((x.size(0),self.num_target_classes),dtype=torch.float).cuda(self.device)
        for i in range(self.num_experts):
            out = self.expertBase[i](x)
            #print(x.size())
            out = self.expertDense[i](out.view(out.size(0), -1))
            pred+=out
        return torch.softmax(pred, dim=1)
    # 1. Fix y so that we get a vector output... on everything, then redo all previous tests
    # 2. If 1. didn't work, use already implemented logsumexp
    # 3. If 2. didn't work, try using a squashing function on out to prevent y-out from becoming too large

    def loss(self,x,y):
        x=torch.tensor(x,dtype=torch.float).cuda(self.device)
        #Populating  true hot encoding
        true=torch.zeros((x.size()[0],self.num_target_classes),device=self.device)
        for i, target in enumerate(y):
            true[i,int(target)]=1.
        #Calculating output from model
        res=torch.zeros((self.num_experts,x.size()[0],self.num_target_classes),dtype=torch.float,device=self.device)
        for i in range(self.num_experts):
            #Getting output from each expert
            _out=self.expertBase[i](x)
            res[i]=self.expertDense[i](_out.view(_out.size(0), -1))
        # calculate correlation
        total_out=torch.mean(res,dim=0)
        correlations=torch.zeros((self.num_experts,x.size()[0]),device=self.device)
        _diff=torch.norm(res-total_out,dim=2)

        for i in range(correlations.size()[0]):
            expert_to_expert=torch.zeros(x.size()[0],device=self.device)
            for j, diff in enumerate(_diff):
                if i!=j:
                    expert_to_expert+=diff
            correlations[i]=torch.norm(res[i]-total_out,dim=1)*expert_to_expert
        #Calculate individual error
        individual_loss=torch.zeros(self.num_experts,device=self.device)
        for i,expert_out in enumerate(res):
            individual_loss[i]=0.5*(torch.norm(expert_out-true)**2)
        #print(individual_loss.size())
        #print(correlations.size())
        loss=torch.mean(individual_loss+(self.penalty_strength*correlations.transpose(0,1)))
        #combine outputs and calculate loss
        #print(loss) 
        return loss 

    def accuracy(self,gen_test_batch,forgiveness=1):
        torch.cuda.empty_cache()
        accuracy=[]
        print('Calculating accuracy')
        for data,target in gen_test_batch:
            data=torch.tensor(data,dtype=torch.float).cuda(self.device)
            target = torch.tensor(target, dtype=torch.float).cuda(self.device)
            accuracy.append(torch.mean(torch.tensor(
                [target[i] in batch_item for i,batch_item in enumerate(self.f(data).topk(k=forgiveness,dim=1).indices)],
                dtype=torch.float
            )))
        out = np.average(accuracy)
        print('Current accuracy %f'%out)
        return out

    def load(self):
        if os.path.isfile('Models/NegativeCorrelationLearning_M1_state.pt'):
            state = torch.load('Models/NegativeCorrelationLearning_M1_state.pt')
            self.load_state_dict(state['model_state'])
            self.optimizer.load_state_dict(state['optimizer_state'])

    def save(self):
        print("Saving...")
        self.to(torch.device('cpu'))
        self.optimizer.zero_grad()
        torch.save({
            'model_state': self.state_dict(),
            'optimizer_state': self.optimizer.state_dict()
        }, 'Models/NegativeCorrelationLearning_M1_state.pt')
        self.to(self.device)

    @staticmethod
    def clear():
        print("Clearing unused variables")
        torch.cuda.empty_cache()


    def trainer(self, gen_training_batch,gen_test_batch,batch_size):
        print("Training on %s"%self.device)
        loss=np.zeros(self.max_epoch)
        accuracy=np.zeros(self.max_epoch)
        for epoch in range(self.max_epoch):
            # Loss
            loss_epoch=[]
            print('Starting training epoch %d'%epoch)
            for data, target in gen_training_batch[epoch]:
                self.optimizer.zero_grad()
                current_loss =self.loss(data.reshape((batch_size,1,self.input_shape[0],self.input_shape[1])),target)
                current_loss.backward()
                self.optimizer.step()
                loss_epoch.append(current_loss.item())
            
            accuracy[epoch]=self.accuracy(gen_test_batch[epoch])
            loss[epoch]=np.average(loss_epoch)
            print('Finishing trainng: Average loss: %f\n'%loss[epoch])
        return loss,accuracy


