import torch
import torch.nn as nn
import numpy as np
import os

class VanillaCNNM4(nn.Module):
    device=torch.device("cuda"if torch.cuda.is_available() else "cpu")

    def __init__(self,input_shape,num_target_classes):
        super(VanillaCNNM4,self).__init__()
        # Based on M7-1
        depth=64
        dense_depth=4096
        self.validation=[]
        self.max_epoch=5
        self.input_shape=input_shape
        self.num_target_classes=num_target_classes
        self.CEloss=nn.CrossEntropyLoss()
        self.CNN_model=nn.Sequential(
            nn.Conv2d(1,depth,kernel_size=3,padding=1),
            nn.BatchNorm2d(depth),
            nn.PReLU(),
            nn.Conv2d(depth,depth,kernel_size=3,padding=1),
            nn.BatchNorm2d(depth),
            nn.PReLU(),
            nn.MaxPool2d(2),

            nn.Conv2d(depth, depth*2, kernel_size=3, padding=1),
            nn.BatchNorm2d(depth*2),
            nn.PReLU(),
            nn.Conv2d(depth*2, depth*2, kernel_size=3, padding=1),
            nn.BatchNorm2d(depth*2),
            nn.PReLU(),
            nn.MaxPool2d(2),

            nn.Conv2d(depth*2, depth*4, kernel_size=3, padding=1),
            nn.BatchNorm2d(depth*4),
            nn.PReLU(),
            nn.Conv2d(depth*4, depth * 4, kernel_size=3, padding=1),
            nn.BatchNorm2d(depth*4),
            nn.PReLU(),
            nn.Conv2d(depth*4, depth * 4, kernel_size=3, padding=1),
            nn.BatchNorm2d(depth*4),
            nn.PReLU(),
            nn.MaxPool2d(2),
            
            nn.Conv2d(depth*4, depth*8, kernel_size=3, padding=1),
            nn.BatchNorm2d(depth*8),
            nn.PReLU(),
            nn.Conv2d(depth*8, depth * 8, kernel_size=3, padding=1),
            nn.BatchNorm2d(depth*8),
            nn.PReLU(),
            nn.Conv2d(depth*8, depth * 8, kernel_size=3, padding=1),
            nn.BatchNorm2d(depth*8),
            nn.PReLU(),
            nn.MaxPool2d(2),
            
            nn.Conv2d(depth*8, depth*8, kernel_size=3, padding=1),
            nn.BatchNorm2d(depth*8),
            nn.PReLU(),
            nn.Conv2d(depth*8, depth * 8, kernel_size=3, padding=1),
            nn.BatchNorm2d(depth*8),
            nn.PReLU(),
            nn.Conv2d(depth*8, depth * 8, kernel_size=3, padding=1),
            nn.BatchNorm2d(depth*8),
            nn.PReLU(),
            nn.MaxPool2d(2))

        self.dense_model=nn.Sequential(
            nn.Linear(depth*2*2*8,dense_depth),
            nn.PReLU(),
            nn.Linear(dense_depth,dense_depth),
            nn.PReLU(),
            nn.Linear(dense_depth,num_target_classes)
        )

        self.optimizer=torch.optim.Adam(self.parameters(),float('1.0e-4'))

    def f(self, x):
        x = x.cuda(self.device)
        x = self.CNN_model(x.reshape((-1, 1, self.input_shape[0], self.input_shape[1])))

        x = self.dense_model(x.view(x.size(0), -1))
        return torch.softmax(x, dim=1)

    def loss(self,x,y):
        x=torch.tensor(x,dtype=torch.float).cuda(self.device)
        y=torch.tensor(y,dtype=torch.long).cuda(self.device)
        x = self.CNN_model(x.reshape((-1, 1, self.input_shape[0], self.input_shape[1])))
        #print(x.size())
        x = self.dense_model(x.view(x.size(0), -1))
        #print(y.size())
        #print(x.size())
        return self.CEloss(x, y)

    def accuracy(self,gen_test_batch,forgiveness=1):
        torch.cuda.empty_cache()
        accuracy=[]
        print('Calculating accuracy')
        for data,target in gen_test_batch:
            data=torch.tensor(data,dtype=torch.float).cuda(self.device)
            target = torch.tensor(target, dtype=torch.float).cuda(self.device)
            accuracy.append(torch.mean(torch.tensor(
                [target[i] in batch_item for i,batch_item in enumerate(self.f(data).topk(k=forgiveness,dim=1).indices)],
                dtype=torch.float
            )))
        out = np.average(accuracy)
        print('Current accuracy %f'%out)
        return out

    def load(self):
        if os.path.isfile('Models/Vanilla_CNN_M4_state.pt'):
            state = torch.load('Models/Vanilla_CNN_M4_state.pt')
            self.load_state_dict(state['model_state'])
            self.optimizer.load_state_dict(state['optimizer_state'])

    def save(self):
        print("Saving...")
        self.to(torch.device('cpu'))
        torch.save({
            'model_state':self.state_dict(),
            'optimizer_state':self.optimizer.state_dict()
        },'Models/Vanilla_CNN_M4_state.pt')
        self.to(self.device)

    @staticmethod
    def clear():
        print("Clearing unused variables")
        torch.cuda.empty_cache()

    def trainer(self, gen_training_batch,gen_test_batch,batch_size):
        print("Training on %s"%self.device)
        loss=np.zeros(self.max_epoch)
        accuracy=np.zeros(self.max_epoch)
        for epoch in range(self.max_epoch):
            # Loss
            loss_epoch=[]
            print('Starting training epoch %d'%epoch)
            for data, target in gen_training_batch[epoch]:
                self.optimizer.zero_grad()
                current_loss =self.loss(data.reshape((batch_size,1,self.input_shape[0],self.input_shape[1])),target)
                current_loss.backward()
                self.optimizer.step()
                loss_epoch.append(current_loss.item())
            accuracy[epoch]=self.accuracy(gen_test_batch[epoch])
            loss[epoch]=np.average(loss_epoch)
            print('Finishing training: Average loss: %f\n'%loss[epoch])
        return loss,accuracy

